use std::collections::HashMap;
use std::sync::Arc;
use lazy_static::lazy_static;
use stretto::Cache;
use tokio::sync::{Mutex, RwLock};
use diesel::prelude::*;
use crate::db::Inode;

lazy_static! {
    pub(super) static ref DB_MANAGER: DBManager = DBManager::new();
}

pub(super) struct DBManager {
    locks: Mutex<HashMap<i32, Arc<RwLock<()>>>>,
    node_cache: Cache<i32, Inode>,
    children_cache: Cache<i32, Vec<Inode>>
}

impl DBManager {
    fn new() -> Self {
        Self {
            locks: Mutex::new(HashMap::new()),
            node_cache: Cache::new(10000, 1000).expect("Failed to create node cache"),
            children_cache: Cache::new(1000, 100).expect("Failed to create child cache")
        }
    }

    pub fn create_node(&self, db: &mut super::RawDBConnection, file: bool, _name: String, parent: Option<i32>, owner: i32) -> Inode {
        use crate::schema::inode::dsl::*;
        let node: Inode = diesel::insert_into(inode)
            .values(crate::db::inode::NewInode {
                is_file: file,
                name: _name,
                parent_id: parent,
                owner_id: owner,
                size: None,
                has_preview: false
            })
            .get_result(db)
            .expect("Failed to insert new inode");

        self.node_cache.insert(node.id, node.clone(), 1);
        if let Some(parent) = parent {
            self.children_cache.remove(&parent);
        }

        node
    }

    pub fn get_node(&self, db: &mut super::RawDBConnection, node_id: i32) -> Option<Inode> {
        use crate::schema::inode::dsl::*;
        let node = self.node_cache.get(&node_id);
        match node {
            Some(v) => Some(v.value().clone()),
            None => {
                let v: Inode = inode.find(node_id).first(db).ok()?;
                self.node_cache.insert(node_id, v.clone(), 1);
                Some(v)
            }
        }
    }

    pub fn get_children(&self, db: &mut super::RawDBConnection, node_id: i32) -> Vec<Inode> {
        use crate::schema::inode::dsl::*;
        let children = self.children_cache.get(&node_id);
        match children {
            Some(v) => v.value().clone(),
            None => {
                let v = inode.filter(parent_id.eq(node_id)).load(db).expect("Failed to get children of node");
                self.children_cache.insert(node_id, v.clone(), 1);
                v
            }
        }
    }

    pub fn save_node(&self, db: &mut super::RawDBConnection, node: &Inode) {
        self.node_cache.insert(node.id, node.clone(), 1);
        diesel::update(node)
            .set(node.clone())
            .execute(db)
            .expect("Failed to save node");
    }

    pub fn delete_node(&self, db: &mut super::RawDBConnection, node: &Inode) {
        if node.is_file {
            let file_name = format!("./files/{}", node.id);
            let file = std::path::Path::new(&file_name);
            let preview_name = format!("./files/{}_preview.jpg", node.id);
            let preview = std::path::Path::new(&preview_name);
            if file.exists() {
                std::fs::remove_file(file).expect("Failed to delete file");
            }
            if preview.exists() {
                std::fs::remove_file(preview).expect("Failed to delete preview");
            }
        }
        diesel::delete(node).execute(db).expect("Failed to delete node");
        self.node_cache.remove(&node.id);
        self.children_cache.remove(&node.id);
        if let Some(p) = node.parent_id { self.children_cache.remove(&p); }
    }

    pub async fn get_lock(&self, user: i32) -> Arc<RwLock<()>> {
        self.locks.lock().await
            .entry(user)
            .or_insert_with(|| Arc::new(RwLock::new(())))
            .clone()
    }
}
