mod inode;
mod token;
mod user;
pub mod manager;
mod connection;

use diesel::connection::SimpleConnection;
use diesel::sqlite::SqliteConnection;
use diesel::r2d2::{ConnectionManager, Pool, PooledConnection};
use diesel_migrations::{embed_migrations, EmbeddedMigrations, MigrationHarness};
use warp::Filter;

pub use inode::Inode;
pub use token::Token;
pub use user::{User, TfaTypes, UserRole};
use crate::routes::AppError;

type RawDBConnection = PooledConnection<ConnectionManager<SqliteConnection>>;
pub type DBPool = Pool<ConnectionManager<SqliteConnection>>;
pub use connection::DBConnection;

pub const MIGRATIONS: EmbeddedMigrations = embed_migrations!("./migrations");

#[derive(Debug)]
pub struct ConnectionOptions {}

impl diesel::r2d2::CustomizeConnection<SqliteConnection, diesel::r2d2::Error> for ConnectionOptions {
    fn on_acquire(&self, conn: &mut SqliteConnection) -> Result<(), diesel::r2d2::Error> {
        conn.batch_execute("PRAGMA journal_mode = WAL; PRAGMA synchronous = NORMAL; PRAGMA busy_timeout = 15000;")
            .map_err(diesel::r2d2::Error::QueryError)
    }
}

pub fn build_pool() -> Pool<ConnectionManager<SqliteConnection>> {
    Pool::builder()
        .connection_customizer(Box::new(ConnectionOptions {}))
        .build(ConnectionManager::<SqliteConnection>::new("sqlite.db"))
        .expect("Failed to open db")
}

pub fn run_migrations(db: &mut RawDBConnection) {
    db.run_pending_migrations(MIGRATIONS).expect("Failed to run migrations");
}

pub fn with_db(pool: DBPool) -> impl Filter<Extract=(DBConnection, ), Error=warp::reject::Rejection> + Clone {
    warp::any()
        .map(move || pool.clone())
        .and_then(|pool: DBPool| async move {
            match pool.get() {
                Ok(v) => Ok(DBConnection::from(v)),
                Err(_) => AppError::InternalError("Failed to get a database connection").err()
            }
        })
}
