use diesel::backend::RawValue;
use diesel::deserialize::{FromSql, FromSqlRow};
use diesel::prelude::*;
use diesel::serialize::{IsNull, Output, ToSql};
use diesel::sql_types::SmallInt;
use diesel::sqlite::Sqlite;
use serde_repr::{Deserialize_repr, Serialize_repr};

#[repr(i16)]
#[derive(Debug, Copy, Clone, Eq, PartialEq, Deserialize_repr, Serialize_repr, FromSqlRow)]
pub enum UserRole {
    Disabled = 0,
    User = 1,
    Admin = 2
}

impl FromSql<SmallInt, Sqlite> for UserRole {
    fn from_sql(bytes: RawValue<'_, Sqlite>) -> diesel::deserialize::Result<Self> {
        match i16::from_sql(bytes)? {
            1 => Ok(UserRole::User),
            2 => Ok(UserRole::Admin),
            _ => Ok(UserRole::Disabled)
        }
    }
}

impl ToSql<SmallInt, Sqlite> for UserRole {
    fn to_sql<'b>(&'b self, out: &mut Output<'b, '_, Sqlite>) -> diesel::serialize::Result {
        let val: i16 = (*self).into();
        out.set_value(val as i32);
        Ok(IsNull::No)
    }
}

impl From<UserRole> for i16 {
    fn from(v: UserRole) -> Self {
        match v {
            UserRole::Disabled => 0,
            UserRole::User => 1,
            UserRole::Admin => 2
        }
    }
}

#[repr(i16)]
#[derive(Debug, Copy, Clone, Eq, PartialEq, Deserialize_repr, Serialize_repr, FromSqlRow)]
pub enum TfaTypes {
    None = 0,
    Email = 1,
    Totp = 2
}

impl FromSql<SmallInt, Sqlite> for TfaTypes {
    fn from_sql(bytes: RawValue<'_, Sqlite>) -> diesel::deserialize::Result<Self> {
        match i16::from_sql(bytes)? {
            1 => Ok(TfaTypes::Email),
            2 => Ok(TfaTypes::Totp),
            _ => Ok(TfaTypes::None)
        }
    }
}

impl ToSql<SmallInt, Sqlite> for TfaTypes {
    fn to_sql<'b>(&'b self, out: &mut Output<'b, '_, Sqlite>) -> diesel::serialize::Result {
        let val: i16 = (*self).into();
        out.set_value(val as i32);
        Ok(IsNull::No)
    }
}

impl From<TfaTypes> for i16 {
    fn from(v: TfaTypes) -> Self {
        match v {
            TfaTypes::None => 0,
            TfaTypes::Email => 1,
            TfaTypes::Totp => 2
        }
    }
}

#[derive(Queryable, Identifiable, Eq, PartialEq, Debug, AsChangeset, Clone)]
#[diesel(table_name = crate::schema::user)]
#[diesel(treat_none_as_null = true)]
pub struct User {
    pub id: i32,
    pub gitlab: bool,
    pub name: String,
    pub password: String,
    #[diesel(serialize_as = i16)]
    pub role: UserRole,
    pub root_id: i32,
    #[diesel(serialize_as = i16)]
    pub tfa_type: TfaTypes,
    pub tfa_secret: Option<Vec<u8>>,
    pub gitlab_at: Option<String>,
    pub gitlab_rt: Option<String>
}

#[derive(Insertable, Debug)]
#[diesel(table_name = crate::schema::user)]
pub struct NewUser {
    pub gitlab: bool,
    pub name: String,
    pub password: String,
    #[diesel(serialize_as = i16)]
    pub role: UserRole,
    pub root_id: i32,
    #[diesel(serialize_as = i16)]
    pub tfa_type: TfaTypes,
    pub tfa_secret: Option<Vec<u8>>,
    pub gitlab_at: Option<String>,
    pub gitlab_rt: Option<String>
}
