use warp::{Filter, Reply};
use crate::db::{DBConnection, DBPool, with_db};
use crate::dto;
use crate::routes::{AppError, get_reply};
use crate::routes::filters::{admin, UserInfo};

pub fn build_routes(db: DBPool) -> impl Filter<Extract = impl Reply, Error = warp::Rejection> + Clone {
    let users = warp::path!("admin" / "users")
        .and(warp::get())
        .and(admin(db.clone()))
        .and(with_db(db.clone()))
        .and_then(users);
    let set_role = warp::path!("admin" / "set_role")
        .and(warp::post())
        .and(warp::body::json())
        .and(admin(db.clone()))
        .and(with_db(db.clone()))
        .and_then(set_role);
    let logout = warp::path!("admin" / "logout")
        .and(warp::post())
        .and(warp::body::json())
        .and(admin(db.clone()))
        .and(with_db(db.clone()))
        .and_then(logout);
    let delete_user = warp::path!("admin" / "delete")
        .and(warp::post())
        .and(warp::body::json())
        .and(admin(db.clone()))
        .and(with_db(db.clone()))
        .and_then(delete_user);
    let disable_2fa = warp::path!("admin" / "disable_2fa")
        .and(warp::post())
        .and(warp::body::json())
        .and(admin(db.clone()))
        .and(with_db(db.clone()))
        .and_then(disable_2fa);
    let is_admin = warp::path!("admin" / "is_admin")
        .and(warp::get())
        .and(admin(db.clone()))
        .and_then(|_| async { get_reply(&dto::responses::Success {
            statusCode: 200
        }) });
    let get_token = warp::path!("admin" / "get_token" / i32)
        .and(warp::get())
        .and(admin(db.clone()))
        .and(with_db(db))
        .and_then(get_token);

    users.or(set_role).or(logout).or(delete_user).or(disable_2fa).or(is_admin).or(get_token)
}

async fn users(_: UserInfo, mut db: DBConnection) -> Result<impl Reply, warp::Rejection> {
    let users = db.get_users();

    let mut res = dto::responses::AdminUsers {
        statusCode: 200,
        users: Vec::new()
    };

    for user in users {
        res.users.push(dto::responses::AdminUsersEntry {
            id: user.id,
            gitlab: user.gitlab,
            name: user.name,
            role: user.role,
            tfaEnabled: user.tfa_type != crate::db::TfaTypes::None
        });
    }

    get_reply(&res)
}

async fn set_role(data: dto::requests::AdminSetRole, _: UserInfo, mut db: DBConnection) -> Result<impl Reply, warp::Rejection> {
    let mut user = db.get_user(data.user)
        .ok_or(AppError::Forbidden("Invalid user"))?;
    user.role = data.role;
    db.save_user(&user);

    get_reply(&dto::responses::Success {
        statusCode: 200
    })
}

async fn logout(data: dto::requests::Admin, _: UserInfo, mut db: DBConnection) -> Result<impl Reply, warp::Rejection> {
    db.delete_all_tokens(data.user);

    get_reply(&dto::responses::Success {
        statusCode: 200
    })
}

async fn delete_user(data: dto::requests::Admin, _: UserInfo, mut db: DBConnection) -> Result<impl Reply, warp::Rejection> {
    let user = db.get_user(data.user)
        .ok_or(AppError::Forbidden("Invalid user"))?;
    
    db.delete_all_tokens(data.user);

    let root_node = super::fs::get_node_and_validate(&user, user.root_id, &mut db).expect("Failed to get root node for deleting");

    super::fs::delete_node_root(&root_node, &mut db);

    db.delete_user(&user);

    get_reply(&dto::responses::Success {
        statusCode: 200
    })
}

async fn disable_2fa(data: dto::requests::Admin, _: UserInfo, mut db: DBConnection) -> Result<impl Reply, warp::Rejection> {
    let mut user = db.get_user(data.user)
        .ok_or(AppError::Forbidden("Invalid user"))?;

    user.tfa_type = crate::db::TfaTypes::None;
    db.save_user(&user);

    get_reply(&dto::responses::Success {
        statusCode: 200
    })
}

async fn get_token(user: i32, _: UserInfo, mut db: DBConnection) -> Result<impl Reply, warp::Rejection> {
    let user = db.get_user(user)
        .ok_or(AppError::Forbidden("Invalid user"))?;

    get_reply(&dto::responses::Login {
        statusCode: 200,
        jwt: super::auth::get_token(&user, &mut db)
    })
}
