use warp::Filter;
use crate::db::{DBConnection, DBPool, with_db};
use crate::db::{TfaTypes, UserRole};
use crate::dto;
use crate::dto::requests::ChangePassword;
use crate::routes::{AppError, get_reply};
use crate::routes::filters::{authenticated, UserInfo};

pub fn build_routes(db: DBPool) -> impl Filter<Extract = impl warp::Reply, Error = warp::Rejection> + Clone {
    let login = warp::path!("auth" / "login")
        .and(warp::post())
        .and(warp::body::json())
        .and(with_db(db.clone()))
        .and_then(login);
    let signup = warp::path!("auth" / "signup")
        .and(warp::post())
        .and(warp::body::json())
        .and(with_db(db.clone()))
        .and_then(signup);
    let refresh = warp::path!("auth" / "refresh")
        .and(warp::post())
        .and(authenticated(db.clone()))
        .and(with_db(db.clone()))
        .and_then(refresh);
    let logout_all = warp::path!("auth" / "logout_all")
        .and(warp::post())
        .and(authenticated(db.clone()))
        .and(with_db(db.clone()))
        .and_then(logout_all);
    let change_password = warp::path!("auth" / "change_password")
        .and(warp::post())
        .and(warp::body::json())
        .and(authenticated(db.clone()))
        .and(with_db(db))
        .and_then(change_password);

    login.or(signup).or(refresh).or(logout_all).or(change_password)
}

async fn login(data: dto::requests::Login, mut db: DBConnection)
    -> Result<impl warp::Reply, warp::Rejection> {
    let user = db.find_user(&data.username, false)
        .ok_or(AppError::Unauthorized("Invalid username or password"))?;

    if !argon2::verify_encoded(user.password.as_str(), data.password.as_bytes()).unwrap_or(false) {
        return AppError::Unauthorized("Invalid username or password").err();
    }

    if user.role == UserRole::Disabled {
        return AppError::Unauthorized("Account is disabled").err();
    }

    if user.tfa_type != TfaTypes::None {
        if let Some(otp) = data.otp {
            if !super::tfa::verify2fa(&user, otp) {
                return AppError::Unauthorized("Incorrect 2fa").err();
            }
        } else {
            if user.tfa_type == TfaTypes::Email { super::tfa::send_2fa_mail(&user); }

            return get_reply(&dto::responses::Success {
                statusCode: 200
            });
        }
    }

    get_reply(&dto::responses::Login {
        statusCode: 200,
        jwt: super::get_token(&user, &mut db)
    })
}

async fn signup(data: dto::requests::SignUp, mut db: DBConnection)
    -> Result<impl warp::Reply, warp::Rejection> {
    if db.find_user(&data.username, false).is_some() {
        return AppError::BadRequest("Username is already taken").err();
    }

    db.create_user_password(data.username, super::hash_password(&data.password));

    get_reply(&dto::responses::Success {
        statusCode: 200
    })
}

async fn refresh(info: UserInfo, mut db: DBConnection) -> Result<impl warp::Reply, warp::Rejection> {
    db.delete_token(info.1.id);

    get_reply(&dto::responses::Login {
        statusCode: 200,
        jwt: super::get_token(&info.0, &mut db)
    })
}

async fn logout_all(info: UserInfo, mut db: DBConnection) -> Result<impl warp::Reply, warp::Rejection> {
    db.delete_all_tokens(info.0.id);

    get_reply(&dto::responses::Success {
        statusCode: 200
    })
}

async fn change_password(data: ChangePassword, mut info: UserInfo, mut db: DBConnection) -> Result<impl warp::Reply, warp::Rejection> {
    if !argon2::verify_encoded(info.0.password.as_str(), data.oldPassword.as_bytes()).unwrap_or(false) {
        return AppError::Unauthorized("Old password is wrong").err();
    }

    info.0.password = super::hash_password(&data.newPassword);
    db.save_user(&info.0);
    db.delete_all_tokens(info.0.id);

    get_reply(&dto::responses::Success {
        statusCode: 200
    })
}
