use cached::proc_macro::cached;
use lazy_static::lazy_static;
use warp::{Filter, Reply};
use crate::config::CONFIG;
use crate::db::{DBConnection, DBPool, with_db};
use crate::routes::AppError;

#[derive(serde::Deserialize, Clone, Debug)]
pub struct GitlabTokens {
    pub access_token: String,
    pub refresh_token: String
}

#[derive(serde::Deserialize, Clone, Debug)]
pub struct GitlabUser {
    pub username: String,
    pub is_admin: bool
}

#[derive(serde::Serialize, serde::Deserialize, Clone, Debug)]
pub struct GitlabCallbackQuery {
    pub code: String
}

lazy_static! {
    static ref REDIRECT_URL: String = CONFIG.gitlab_redirect_url.clone() + "/api/auth/gitlab_callback";
    static ref TOKEN_URL: String = format!("{}/oauth/token", CONFIG.gitlab_api_url.clone());
    static ref USER_URL: String = format!("{}/api/v4/user", CONFIG.gitlab_api_url.clone());
    static ref AUTHORIZE_URL: String = format!("{}/oauth/authorize", CONFIG.gitlab_url.clone());
}

pub fn get_gitlab_token(code_or_token: String, token: bool) -> Option<GitlabTokens> {
    let mut req = ureq::post(&TOKEN_URL)
        .query("redirect_uri", &REDIRECT_URL)
        .query("client_id", &CONFIG.gitlab_id)
        .query("client_secret", &CONFIG.gitlab_secret);
    if token {
        req = req
            .query("refresh_token", &code_or_token)
            .query("grant_type", "refresh_token");
    } else {
        req = req
            .query("code", &code_or_token)
            .query("grant_type", "authorization_code");
    }
    req.call().ok()?.into_json().ok()
}

#[cached(time=300, time_refresh=false, option=true)]
pub fn get_gitlab_user(token: String) -> Option<GitlabUser> {
    ureq::get(&USER_URL)
        .set("Authorization", &format!("Bearer {}", token))
        .call()
        .ok()?
        .into_json().ok()
}

pub fn build_routes(db: DBPool) -> impl Filter<Extract = impl Reply, Error = warp::Rejection> + Clone {
    let gitlab = warp::path!("auth" / "gitlab")
        .and(warp::get())
        .and_then(gitlab);
    let gitlab_callback = warp::path!("auth" / "gitlab_callback")
        .and(warp::get())
        .and(warp::query::query::<GitlabCallbackQuery>())
        .and(with_db(db))
        .and_then(gitlab_callback);

    gitlab.or(gitlab_callback)
}

async fn gitlab() -> Result<impl Reply, warp::Rejection> {
    let uri = format!("{}?redirect_uri={}&client_id={}&scope=read_user&response_type=code", AUTHORIZE_URL.as_str(), REDIRECT_URL.as_str(), CONFIG.gitlab_id);
    Ok(warp::redirect::found(uri.parse::<warp::http::Uri>().expect("Failed to parse gitlab auth uri")))
}

async fn gitlab_callback(code: GitlabCallbackQuery, mut db: DBConnection) -> Result<impl Reply, warp::Rejection> {
    use crate::db::UserRole;

    let tokens = get_gitlab_token(code.code, false).ok_or(AppError::Unauthorized("Invalid code"))?;
    let gitlab_user = get_gitlab_user(tokens.access_token.clone()).ok_or(AppError::Unauthorized("Invalid code"))?;

    let user = db.find_user(&gitlab_user.username, true);

    let user = match user {
        Some(mut v) => {
            v.gitlab_at = Some(tokens.access_token);
            v.gitlab_rt = Some(tokens.refresh_token);
            db.save_user(&v);
            v
        },
        None => {
            db.create_user_gitlab(
                gitlab_user.username,
                if gitlab_user.is_admin { UserRole::Admin } else { UserRole::Disabled },
                tokens.access_token,
                tokens.refresh_token
            )
        }
    };

    if user.role == UserRole::Disabled {
        Ok(warp::reply::html("<!DOCTYPE html><html><h2>Your account is disabled, please contact an admin.<br/><a href=\"/login\">Go to login page</a></h2></html>").into_response())
    } else {
        let uri = format!("/set_token?token={}", super::get_token(&user, &mut db));
        Ok(warp::redirect::found(uri.parse::<warp::http::Uri>().expect("Failed to parse set_token uri")).into_response())
    }
}

