mod basic;
mod tfa;
pub mod gitlab;

use std::ops::Add;
use lazy_static::lazy_static;
use ring::rand;
use ring::rand::SecureRandom;
use warp::Filter;
use crate::db::DBPool;

pub fn build_routes(db: DBPool) -> impl Filter<Extract = impl warp::Reply, Error = warp::Rejection> + Clone {
    SEC_RANDOM.fill(&mut [0; 1]).expect("Failed to init secure random");
    basic::build_routes(db.clone())
        .or(tfa::build_routes(db.clone()))
        .or(gitlab::build_routes(db))
}

#[derive(Debug, serde::Deserialize, serde::Serialize)]
pub struct JWTClaims {
    pub exp: i64,
    pub iat: i64,
    pub jti: i32,
    pub sub: i32
}

pub static JWT_ALGORITHM: jsonwebtoken::Algorithm = jsonwebtoken::Algorithm::HS512;

lazy_static! {
    pub static ref SEC_RANDOM: rand::SystemRandom = rand::SystemRandom::new();
    pub static ref JWT_SECRET: Vec<u8> = get_jwt_secret();
    pub static ref JWT_DECODE_KEY: jsonwebtoken::DecodingKey = jsonwebtoken::DecodingKey::from_secret(JWT_SECRET.as_slice());
    pub static ref JWT_ENCODE_KEY: jsonwebtoken::EncodingKey = jsonwebtoken::EncodingKey::from_secret(JWT_SECRET.as_slice());
}

fn get_jwt_secret() -> Vec<u8> {
    let secret = std::fs::read("jwt.secret");
    if let Ok(secret) = secret {
        secret
    } else {
        let mut secret: [u8; 128] = [0; 128];
        SEC_RANDOM.fill(&mut secret).expect("Failed to generate jwt secret");
        std::fs::write("jwt.secret", secret).expect("Failed to write jwt secret");
        Vec::from(secret)
    }
}

pub fn get_token(user: &crate::db::User, db: &mut crate::db::DBConnection) -> String {
    let iat = chrono::Utc::now();
    let exp = iat.add(chrono::Duration::hours(24)).timestamp();
    let iat = iat.timestamp();

    let token = db.create_token(user.id, exp);

    let claims = JWTClaims {
        exp,
        iat,
        jti: token.id,
        sub: user.id
    };

    jsonwebtoken::encode(&jsonwebtoken::Header::new(JWT_ALGORITHM), &claims, &JWT_ENCODE_KEY)
        .expect("Failed to create JWT token")
}

pub fn hash_password(password: &String) -> String {
    let mut salt = [0_u8; 16];
    SEC_RANDOM.fill(&mut salt).expect("Failed to generate salt");
    let config = argon2::Config {
        mem_cost: 64 * 1024,
        variant: argon2::Variant::Argon2id,
        ..Default::default()
    };
    argon2::hash_encoded(password.as_bytes(), &salt, &config).expect("Failed to hash password")
}
