use warp::Filter;
use warp::http::{HeaderMap, HeaderValue};
use crate::db::UserRole;
use crate::db::{DBConnection, DBPool, with_db};
use crate::routes::AppError;
use crate::routes::auth;

pub type UserInfo = (crate::db::User, crate::db::Token);

pub fn authenticated(db: DBPool) -> impl Filter<Extract=(UserInfo,), Error=warp::reject::Rejection> + Clone {
    warp::header::headers_cloned()
        .map(move |_headers: HeaderMap<HeaderValue>| _headers)
        .and(with_db(db))
        .and_then(authorize)
}

pub fn admin(db: DBPool) -> impl Filter<Extract=(UserInfo, ), Error=warp::reject::Rejection> + Clone {
    warp::header::headers_cloned()
        .map(move |_headers: HeaderMap<HeaderValue>| _headers)
        .and(with_db(db))
        .and_then(|_headers, db| async {
        let info = authorize(_headers, db).await?;
        if info.0.role == UserRole::Admin {
            Ok(info)
        } else {
            AppError::Forbidden("Forbidden").err()
        }
    })
}

async fn authorize(_headers: HeaderMap<HeaderValue>, mut db: DBConnection) -> Result<UserInfo, warp::reject::Rejection> {
    authorize_jwt(extract_jwt(&_headers).map_err(|e| e.reject())?, &mut db).await
}

pub async fn authorize_jwt(jwt: String, db: &mut DBConnection) -> Result<UserInfo, warp::reject::Rejection> {
    let decoded = jsonwebtoken::decode::<auth::JWTClaims>(
        &jwt,
        &crate::routes::auth::JWT_DECODE_KEY,
        &jsonwebtoken::Validation::new(auth::JWT_ALGORITHM)
    ).map_err(|_| AppError::Forbidden("Invalid token"))?;

    db.cleanup_tokens();

    let mut user = db.get_user(decoded.claims.sub)
        .ok_or(AppError::Forbidden("Invalid token"))?;
    let token = db.get_token(decoded.claims.jti)
        .ok_or(AppError::Forbidden("Invalid token"))?;

    if user.id != token.owner_id {
        return AppError::Forbidden("Invalid token").err();
    }
    if user.role == UserRole::Disabled {
        return AppError::Forbidden("Account disabled").err();
    }
    if user.gitlab {
        let info = auth::gitlab::get_gitlab_user(user.gitlab_at.clone().unwrap());
        let info = match info {
            Some(v) => Some(v),
            None => {
                let tokens = auth::gitlab::get_gitlab_token(user.gitlab_rt.clone().unwrap(), true);
                if let Some(tokens) = tokens {
                    user.gitlab_at = Some(tokens.access_token.clone());
                    user.gitlab_rt = Some(tokens.refresh_token);
                    db.save_user(&user);
                    auth::gitlab::get_gitlab_user(tokens.access_token)
                } else { None }
            }
        };
        if info.is_none() || info.unwrap().username != user.name {
            db.delete_all_tokens(token.owner_id);
            db.delete_all_tokens(user.id);
            return AppError::Forbidden("Invalid gitlab user").err();
        }
    }

    Ok((user, token))
}

fn extract_jwt(_headers: &HeaderMap<HeaderValue>) -> Result<String, AppError> {
    let header = match _headers.get(warp::http::header::AUTHORIZATION) {
        Some(v) => v,
        None => return Err(AppError::Unauthorized("Missing token"))
    };
    let header = header.to_str().map_err(|_| AppError::Unauthorized("Missing token"))?;
    if !header.starts_with("Bearer ") {
        Err(AppError::Unauthorized("Missing token"))
    } else {
        Ok(header.trim_start_matches("Bearer ").to_owned())
    }
}