mod filters;
mod auth;
mod admin;
mod user;
pub mod fs;

use warp::{Filter, Reply};
use crate::db::DBPool;
use crate::dto;

pub fn build_routes(db: DBPool) -> impl Filter<Extract = impl Reply, Error = warp::Rejection> + Clone {
    warp::path::path("api")
        .and(
            auth::build_routes(db.clone())
                .or(admin::build_routes(db.clone()))
                .or(user::build_routes(db.clone()))
                .or(fs::build_routes(db))
                .recover(error_handler)
        )
        .or(warp::fs::dir("./static/"))
        .or(warp::fs::file("./static/index.html"))
}

pub fn get_reply<T>(data: &T) -> Result<warp::reply::Response, warp::Rejection> where T: serde::Serialize {
    Ok(warp::reply::with_status(warp::reply::json(data), warp::http::StatusCode::OK).into_response())
}

#[derive(thiserror::Error, Debug, Clone)]
pub enum AppError {
    #[error("unauthorized")]
    Unauthorized(&'static str),
    #[error("forbidden")]
    Forbidden(&'static str),
    #[error("bad request")]
    BadRequest(&'static str),
    #[error("internal error")]
    InternalError(&'static str)
}
impl warp::reject::Reject for AppError {}

impl AppError {
    pub fn reject(&self) -> warp::reject::Rejection {
        warp::reject::custom(self.clone())
    }

    pub fn err<T>(&self) -> Result<T, warp::reject::Rejection> {
        Err(self.reject())
    }
}

pub async fn error_handler(err: warp::reject::Rejection) -> Result<impl Reply, std::convert::Infallible> {
    if err.is_not_found() {
        return Ok(warp::reply::with_status(
            warp::reply::json(&dto::responses::Error {
                statusCode: 404,
                message: "bruh".to_owned()
            }),
            warp::http::StatusCode::NOT_FOUND
        ));
    }
    if let Some(e) = err.find::<AppError>() {
        return Ok(warp::reply::with_status(
            warp::reply::json(&dto::responses::Error {
                statusCode: match e {
                    AppError::BadRequest(_) => 400,
                    AppError::Unauthorized(_) => 401,
                    AppError::Forbidden(_) => 403,
                    AppError::InternalError(_) => 500
                },
                message: match e {
                    AppError::BadRequest(v) => v.to_string(),
                    AppError::Unauthorized(v) => v.to_string(),
                    AppError::Forbidden(v) => v.to_string(),
                    AppError::InternalError(v) => v.to_string()
                },
            }),
            match e {
                AppError::BadRequest(_) => warp::http::StatusCode::BAD_REQUEST,
                AppError::Unauthorized(_) => warp::http::StatusCode::UNAUTHORIZED,
                AppError::Forbidden(_) => warp::http::StatusCode::FORBIDDEN,
                AppError::InternalError(_) => warp::http::StatusCode::INTERNAL_SERVER_ERROR
            }
        ));
    }
    if let Some(e) = err.find::<warp::body::BodyDeserializeError>() {
        return Ok(warp::reply::with_status(
            warp::reply::json(&dto::responses::Error {
                statusCode: 400,
                message: e.to_string(),
            }),
            warp::http::StatusCode::BAD_REQUEST
        ))
    }
    if let Some(e) = err.find::<warp::reject::InvalidQuery>() {
        return Ok(warp::reply::with_status(
            warp::reply::json(&dto::responses::Error {
                statusCode: 400,
                message: e.to_string(),
            }),
            warp::http::StatusCode::BAD_REQUEST
        ))
    }
    if let Some(e) = err.find::<warp::reject::MethodNotAllowed>() {
        return Ok(warp::reply::with_status(
            warp::reply::json(&dto::responses::Error {
                statusCode: 405,
                message: e.to_string(),
            }),
            warp::http::StatusCode::METHOD_NOT_ALLOWED
        ))
    }

    Err(err).expect("Can't handle error")
}